/* 
  Copyright (C) 2009 Bengt Martensson <barf@bengt-martensson.de>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

 */

package com.hifiremote.decodeir;

/**
 * This is a wrapper arount the DecodeIDCaller class, intended to look
 * more Java-like.
 */

public class DecodeIR {
    public static class DecodedSignal {
	public String protocolName;
	public int device;
	public int subDevice;
	public int OBC;
	public int[] hex;
	public String miscMessage;
	public String errorMessage;

	public DecodedSignal(String protocolName,
			     int device,
			     int subDevice,
			     int OBC,
			     int[] hex,
			     String miscMessage,
			     String errorMessage) {
	    this.protocolName = protocolName;
	    this.device = device;
	    this.subDevice = subDevice;
	    this.OBC = OBC;
	    this.hex = hex;
	    this.miscMessage = miscMessage;
	    this.errorMessage = errorMessage;
	}

	public DecodedSignal(String protocolName,
			     int device,
			     int subDevice,
			     int OBC) {
	    this(protocolName, device, subDevice, OBC, null, "", "");
	}

	public DecodedSignal(String protocolName,
			     int device,
			     int OBC) {
	    this(protocolName, device, -1, OBC);
	}


	public String toString() {
	    return "protocol = " + protocolName
		+ ", device = " + device
		+ ((subDevice >= 0) ? (", subdevice = " + subDevice) : "")
		+ ", obc = " + OBC
		//+ ", hex = " + getHex
		+ (!miscMessage.equals("") ? (", misc = " + miscMessage) : "")
		+ (!errorMessage.equals("") ? (", error = " + errorMessage) : "");
	}
    }

    private DecodeIRCaller dirc;
    private int no_decodes;
    private final static int max_decodes = 10;
    private DecodedSignal[] work;
    private DecodedSignal[] decodedSignals;

    public static String getVersion() {
	DecodeIRCaller d = new DecodeIRCaller();
	return d.getVersion();
    }

    private void setup(int[] data, int length_repeat, int frequency) {
	//System.err.println(frequency + ", " + length_repeat + ", " + data.length);
	//for (int i = 0; i < data.length; i++)
	//    System.err.println(data[i]);
	int[] us_data = new int[data.length];
	for (int i = 0; i < data.length; i++) {
	    us_data[i] = (int)(1000000.0/frequency*data[i]);
	    //System.err.println(us_data[i]);
	}
	System.err.println(frequency);

	dirc = new DecodeIRCaller();
	dirc.setBursts(us_data, length_repeat);
	dirc.setFrequency(frequency);
	dirc.initDecoder();
	
	work = new DecodedSignal[max_decodes];
	int no_decodes = 0;
	try {
	    while (dirc.decode()) {
		work[no_decodes] = new DecodedSignal(dirc.getProtocolName(),
						     dirc.getDevice(),
						     dirc.getSubDevice(),
						     dirc.getOBC(),
						     dirc.getHex(),
						     dirc.getMiscMessage(),
						     dirc.getErrorMessage());
		no_decodes++;
	    };
	} catch (ArrayIndexOutOfBoundsException e) {
	    System.err.println("Only considering the first " + max_decodes + " decodes.");
	}

	decodedSignals = new DecodedSignal[no_decodes];
	for (int i = 0; i < no_decodes; i++)
	    decodedSignals[i] = work[i];
    }

    public DecodeIR(int[] data, int length_repeat, int frequency) {
	setup(data, length_repeat, frequency);
    }

    public DecodeIR(int[] CCF) {
	int i = 0;
	if (CCF[i++] != 0) {
	    System.err.println("Can only handle type 0000");
	    return;
	}
	    
	int fcode = CCF[i++];
	int freq = (int) (1000000.0/((double)fcode * 0.241246));
	int intro_length = CCF[i++];
	int rep_length = CCF[i++];

	int[] data = new int[2*(intro_length+rep_length)];
	for (int j = 0; j < 2*(intro_length+rep_length); j++)
	    data[j] = CCF[i++];
	setup(data, rep_length, freq);
    }
 
    public DecodeIR(String ccf) {
	this(decodeCCF(ccf));
    }
 
    private static int[] decodeCCF(String s) throws NumberFormatException {
	String[] work = s.split("[ \t\n]+");
	int[] result = new int[work.length];
	    for (int i = 0; i < result.length; i++)
		result[i] = Integer.parseInt(work[i], 16);

	return result;
    }

    public DecodedSignal[] getDecodedSignals() {
	return decodedSignals;
    }

    public static DecodedSignal[] decode(int[] data, int length_repeat, int frequency) {
	return (new DecodeIR(data, length_repeat, frequency).getDecodedSignals());
    }

    public static DecodedSignal[] decode(int[] ccf) {
	DecodedSignal[] result = null;
	switch (ccf[0]) {
	case 0:
	    result = (new DecodeIR(ccf)).getDecodedSignals();
	    break;
	case 0x5000:
	    if (ccf[2] == 0 && ccf[3] == 1) {
		result = new DecodedSignal[1];
		result[0] = new DecodedSignal("RC5", ccf[4], ccf[5]);
	    }
	    break;
	case 0x5001:
	    if (ccf[2] == 0 && ccf[3] == 2) {
		result = new DecodedSignal[1];
		result[0] = new DecodedSignal("RC5x", ccf[4], ccf[5], ccf[6]);
	    }
	    break;
	case 0x900a:
	    if (ccf[2] == 0 && ccf[3] == 1) {
		int dev = ccf[4] >> 8;
                int subdev = ccf[4] & 0xff;
		int cmd = ccf[5] >> 8;
		int cmd_chk = 0xff - (ccf[5] & 0xff);
		if (cmd == cmd_chk) {
		    result = new DecodedSignal[1];
		    result[0] = dev == 255-subdev ? new DecodedSignal("NEC1", dev, cmd) : new DecodedSignal("NEC1", dev, subdev, cmd);
		}
	    }
	    break;
	default:
	    System.err.println("Unknown CCF type: " + Integer.toHexString(ccf[0]));
	}
	return result;
    }

    public static DecodedSignal[] decode(String ccf) {
	return decode(decodeCCF(ccf));
    }

    private static void usage() {
	System.err.println("Usage:");
	System.err.println("DecodeIR [-v] CCF-code.");
	System.exit(0);
    }

    public static void main(String[] args) {
	boolean show_version = false;

	try {
	    int arg_i = 0;
	    while (arg_i < args.length
		   && (args[arg_i].length() > 0)
		   && args[arg_i].charAt(0) == '-') {

		if (args[arg_i].equals("-v")) {
		    arg_i++;
		    show_version = true;
		} else
		    usage();
	    }
	    if (show_version) {
		System.out.println("DecodeIR version " + DecodeIR.getVersion());
		System.exit(0);
	    }
	    
	    DecodedSignal[] result;
	    if (args.length == arg_i + 1) {
		result = decode(args[arg_i]);
	    } else {
		int[] ccf = new int[args.length - arg_i];
		for (int i = 0; i < args.length - arg_i; i++)
		    ccf[i] = Integer.parseInt(args[arg_i+i], 16);
		result = decode(ccf);
	    }
	    if (result != null)
		for (int i = 0; i < result.length; i++)
		    System.out.println(result[i]);

       	} catch (ArrayIndexOutOfBoundsException e) {
	    //System.err.println("ArrayIndexOutOfBoundsException");
	    usage();
	    //} catch (NumberFormatException e) {
	    //System.err.println("NumberFormatException");
	    //usage();
	}

    }
}
